/*
   Per-Vertex Lighting Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform per-vertex diffuse lighting.  To do this we will
   calculate the lighting color per vertex.  The lighting equation we will use is this...

   final color = ambient final color + diffuse final color
   
   To calculate the ambient color we use...

   ambient final color = ambient color * ambient material color

   And to calculate the diffuse color we use...

   diffuse final color = dot product of(light pos - vertex pos) * light color *
                         object diffuse material color

   This is all done in the vertex shader.
*/


// Demo: Per-Vertex Lighting.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>

// Our vertex stucture define.
#define D3DFVF_D3DVertex (D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1)

// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
void ShutdownDirect3D();

// Here are our Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;
LPDIRECT3DVERTEXBUFFER9 wall1Buffer = NULL;
LPDIRECT3DVERTEXBUFFER9 wall2Buffer = NULL;
LPDIRECT3DVERTEXBUFFER9 wall3Buffer = NULL;
LPDIRECT3DVERTEXBUFFER9 wall4Buffer = NULL;
LPDIRECT3DVERTEXBUFFER9 wall5Buffer = NULL;
LPDIRECT3DTEXTURE9 roomTexture = NULL;

LPD3DXMESH sphereMesh;

LPD3DXEFFECT effectShader = NULL;

// Projection and mvp matrix.
D3DXMATRIXA16 Projection;
D3DXMATRIX modelViewProjection;
D3DXMATRIX modelView;

// Light information.
float lightPos[4] = { 0.0f, 0.0f, 1.0f, 1.0f };
float lightColor[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
float ambientMat[4] = { 0.6f, 0.6f, 0.6f, 1.0f };
float ambientCol[4] = { 0.6f, 0.6f, 0.6f, 1.0f };
float diffuseMat[4] = { 0.8f, 0.8f, 0.8f, 1.0f };

// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
   FLOAT nx, ny, nz;
   FLOAT tu, tv;
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true if we should exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Per-Vertex Lighting - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           D3DCREATE_HARDWARE_VERTEXPROCESSING,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // Here we set up the object this demo will render.
   if(!InitializeObject()) return false;

   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);       // Lighting on.
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);// Culling off.
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);         // Enable depth testing.

   // We can use this to store any errors when compiling the shader effect file.
	LPD3DXBUFFER errors = NULL;

   // Load the effect file with all of our code.
	if(FAILED(D3DXCreateEffectFromFile(D3D_Device, "effect.fx", NULL, NULL, 0,
                                      NULL, &effectShader, &errors)))
      {
		   LPVOID compileErrors = errors->GetBufferPointer();
		   
         // Show the errors to the user.
         MessageBox(NULL, (const char*)compileErrors, "Shader Compile Error...",
                    MB_OK | MB_ICONEXCLAMATION);

         return false;
	   }

   // Load the roomTexture image.
   if(FAILED(D3DXCreateTextureFromFile(D3D_Device, "ugp.bmp", &roomTexture)))
      return false;

   // Set the roomTexture min/mag info.
   D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   // Set the perspective information.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, (float)(640 / 480),
                              0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);

   return true;
}


bool InitializeObject()
{
   // Square data for its position, roomTexture coords, and vertex color.
	Vertex wall1[] =//back
	{
		{-1,  1, 1, 0, 0, -1, 0, 0}, { 1,  1, 1, 0, 0, -1, 1, 0},
	   {-1, -1, 1, 0, 0, -1, 0, 1}, {-1, -1, 1, 0, 0, -1, 0, 1},
	   { 1, -1, 1, 0, 0, -1, 1, 1}, { 1,  1, 1, 0, 0, -1, 1, 0}
	};
	
	Vertex wall2[] =// roof
	{
	   {-1, 1,  1, 0, -1, 0, 0, 0}, { 1, 1,  1, 0, -1, 0, 1, 0},
	   {-1, 1, -1, 0, -1, 0, 0, 1}, {-1, 1, -1, 0, -1, 0, 0, 1},
	   { 1, 1, -1, 0, -1, 0, 1, 1}, { 1, 1,  1, 0, -1, 0, 1, 0}
	};
	
	Vertex wall3[] =// left
	{
	   {-1,  1,  1, 1, 0, 0, 1, 0}, {-1, -1,  1, 1, 0, 0, 1, 1},
	   {-1,  1, -1, 1, 0, 0, 0, 0}, {-1,  1, -1, 1, 0, 0, 0, 0},
	   {-1, -1, -1, 1, 0, 0, 0, 1}, {-1, -1,  1, 1, 0, 0, 1, 1}
	};
	
	Vertex wall4[] =// right
	{
	   {1,  1,  1, -1, 0, 0, 0, 0}, {1, -1,  1, -1, 0, 0, 0, 1},
	   {1,  1, -1, -1, 0, 0, 1, 0}, {1,  1, -1, -1, 0, 0, 1, 0},
	   {1, -1, -1, -1, 0, 0, 1, 1}, {1, -1,  1, -1, 0, 0, 0, 1}
	};
	
	Vertex wall5[] =// floor
	{
	   {-1, -1,  1, 0, 1, 0, 0, 0}, { 1, -1,  1, 0, 1, 0, 1, 0},
	   {-1, -1, -1, 0, 1, 0, 0, 1}, {-1, -1, -1, 0, 1, 0, 0, 1},
	   { 1, -1, -1, 0, 1, 0, 1, 1}, { 1, -1,  1, 0, 1, 0, 1, 0}
	};

   // Create the vertex buffer that will hold the polygon.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(wall1), 0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &wall1Buffer, NULL)))
      return false;
   
   if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(wall2), 0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &wall2Buffer, NULL)))
      return false;
   
   if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(wall3), 0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &wall3Buffer, NULL)))
      return false;
   
   if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(wall4), 0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &wall4Buffer, NULL)))
      return false;
   
   if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(wall5), 0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &wall5Buffer, NULL)))
      return false;
	
   // Pointer to the buffer.
   Vertex* Vertices;

   if(FAILED(wall1Buffer->Lock(0, sizeof(wall1), (void**)&Vertices, 0)))
      return false;
   memcpy(Vertices, wall1, sizeof(wall1));
   wall1Buffer->Unlock();
   
   if(FAILED(wall2Buffer->Lock(0, sizeof(wall2), (void**)&Vertices, 0)))
      return false;
   memcpy(Vertices, wall2, sizeof(wall2));
   wall2Buffer->Unlock();
   
   if(FAILED(wall3Buffer->Lock(0, sizeof(wall3), (void**)&Vertices, 0)))
      return false;
   memcpy(Vertices, wall3, sizeof(wall3));
   wall3Buffer->Unlock();
   
   if(FAILED(wall4Buffer->Lock(0, sizeof(wall4), (void**)&Vertices, 0)))
      return false;
   memcpy(Vertices, wall4, sizeof(wall4));
   wall4Buffer->Unlock();
   
   if(FAILED(wall5Buffer->Lock(0, sizeof(wall5), (void**)&Vertices, 0)))
      return false;
   memcpy(Vertices, wall5, sizeof(wall5));
   wall5Buffer->Unlock();
   
   if(FAILED(D3DXCreateSphere(D3D_Device, 0.05f, 25, 25, &sphereMesh, NULL)))
      return false;
   
   return true;
}


void RenderScene()
{
   // Clear the screen.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_COLORVALUE(0.0f, 0.0f, 0.0f, 1.0f), 1.0f, 0);

   // This will hold the view matrix.
   D3DXMATRIXA16 View;
   D3DXMatrixIdentity(&View);
   D3DXVECTOR3 Eye(0.0f, 0.0f, -2.8f);
   D3DXVECTOR3 LookAt(0.0f, 0.0f, 0.0f);
   D3DXVECTOR3 Up(0.0f, 1.0f, 0.0f);

   D3DXMatrixLookAtLH(&View, &Eye, &LookAt, &Up);

   modelViewProjection = View * Projection;
	D3DXMatrixTranspose(&modelViewProjection, &modelViewProjection);
	
	modelView = View;
	D3DXMatrixTranspose(&modelView, &modelView);

   D3D_Device->BeginScene();
   
      // Set which technique from our shader file we will use.
      effectShader->SetTechnique("PerVertexLighting");
      effectShader->SetTexture("decal", roomTexture);
      effectShader->SetMatrix("modelViewProj", &modelViewProjection);
      effectShader->SetFloatArray("lightColor", lightColor, 4);
      effectShader->SetFloatArray("lightPos", lightPos, 4);
      effectShader->SetFloatArray("ambientMat", ambientMat, 4);
      effectShader->SetFloatArray("ambientCol", ambientCol, 4);
      effectShader->SetFloatArray("diffuseMat", diffuseMat, 4);

      effectShader->Begin(NULL, 0);

         effectShader->BeginPass(0);

            // Draw the room.
            D3D_Device->SetStreamSource(0, wall1Buffer, 0, sizeof(Vertex));
            D3D_Device->SetFVF(D3DFVF_D3DVertex);
            D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);
            
            D3D_Device->SetStreamSource(0, wall2Buffer, 0, sizeof(Vertex));
            D3D_Device->SetFVF(D3DFVF_D3DVertex);
            D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);
            
            D3D_Device->SetStreamSource(0, wall3Buffer, 0, sizeof(Vertex));
            D3D_Device->SetFVF(D3DFVF_D3DVertex);
            D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);
            
            D3D_Device->SetStreamSource(0, wall4Buffer, 0, sizeof(Vertex));
            D3D_Device->SetFVF(D3DFVF_D3DVertex);
            D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);
            
            D3D_Device->SetStreamSource(0, wall5Buffer, 0, sizeof(Vertex));
            D3D_Device->SetFVF(D3DFVF_D3DVertex);
            D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);

         effectShader->EndPass();

      effectShader->End();
   
      D3DXMATRIXA16 lightPosMat;
      D3DXMatrixTranslation(&lightPosMat, lightPos[0], lightPos[1], lightPos[2]);

      modelViewProjection = lightPosMat * View * Projection;
      D3DXMatrixTranspose(&modelViewProjection, &modelViewProjection);

      effectShader->SetTechnique("SimpleLight");
      effectShader->SetMatrix("modelViewProj", &modelViewProjection);
      effectShader->SetMatrix("modelView", &modelView);
      effectShader->SetFloatArray("lightColor", lightColor, 4);
      effectShader->SetFloatArray("lightPos", lightPos, 4);

      effectShader->Begin(NULL, 0);

         effectShader->BeginPass(0);

            sphereMesh->DrawSubset(0);

         effectShader->EndPass();

      effectShader->End();

   D3D_Device->EndScene();

   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
   
   // Gather user input.
   if(GetKeyState(VK_LEFT) & 0x80) lightPos[0] -= 0.05f;
   if(GetKeyState(VK_RIGHT) & 0x80) lightPos[0] += 0.05f;

   if(GetKeyState(VK_DOWN) & 0x80) lightPos[1] -= 0.05f;
   if(GetKeyState(VK_UP) & 0x80) lightPos[1] += 0.05f;
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(effectShader != NULL)
      {
         effectShader->Release();
         effectShader = NULL;
      }

   if(roomTexture != NULL) 
      {
         roomTexture->Release(); 
         roomTexture = NULL;
      }
   
   if(wall1Buffer != NULL)
      {
         wall1Buffer->Release();
         wall1Buffer = NULL;
      }
   
   if(wall2Buffer != NULL)
      {
         wall2Buffer->Release();
         wall2Buffer = NULL;
      }
   
   if(wall3Buffer != NULL)
      {
         wall3Buffer->Release();
         wall3Buffer = NULL;
      }
   
   if(wall4Buffer != NULL)
      {
         wall4Buffer->Release();
         wall4Buffer = NULL;
      }
   
   if(wall5Buffer != NULL)
      {
         wall5Buffer->Release();
         wall5Buffer = NULL;
      }
   
   if(sphereMesh != NULL)
      {
         sphereMesh->Release();
         sphereMesh = NULL;
      }
}


// Recap:

// After you run this demo you'll see why per-vertex lighting is not the way to
// go.  If the light does not move the scene looks ok but when the light moves you
// can tell that the walls are only made up of 2 triangles.  To get the lighting over
// the surface to look good, we need to perform the lighting equation on a per-pixel
// level.  Either that or you will need to increase the triangle count by a lot.


// Copyright December 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com